;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; classes.lisp -- persistent objects via metaobjects
;;; 
;;; Initial version 8/26/2004 by Andrew Blumberg
;;; <ablumberg@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(in-package "ELEPHANT")

(defmethod initialize-instance :before  ((instance persistent)
					 &rest initargs
					 &key from-oid)
  (declare (ignore initargs))
  "Sets the OID."
  (if (not from-oid)
      (setf (oid instance) (next-oid *store-controller*))
      (setf (oid instance) from-oid))
  (cache-instance *store-controller* instance))

(defclass persistent-object (persistent)
  ((%persistent-slots :transient t))
  (:documentation "Superclass of all user-defined persistent
classes")
  (:metaclass persistent-metaclass))

#|
(defmethod compute-class-precedence-list :around ((class persistent-metaclass))
  (let ((cpl (call-next-method))
	(persistent-object (find-class 'persistent-object)))
    (if (member persistent-object cpl :test #'eq)
	cpl
	(let ((std-obj (find-class 'standard-object))
	      (ccpl (copy-list cpl)))
	  (loop for c on ccpl
		when (eq (cadr c) std-obj)
		do
		(setf (cdr c) (cons persistent-object 
				    (cons (find-class 'persistent) (cdr c))))
		(return nil))
	  ccpl))))
|#

(defmethod shared-initialize :around ((class persistent-metaclass) slot-names &rest args &key direct-superclasses)
  (let* ((persistent-metaclass (find-class 'persistent-metaclass))
	 (persistent-object (find-class 'persistent-object))
	 (not-already-persistent (loop for superclass in direct-superclasses
				       never (eq (class-of superclass) persistent-metaclass))))
    (if (and (not (eq class persistent-object)) not-already-persistent)
	(apply #'call-next-method class slot-names
	       :direct-superclasses (cons persistent-object
					  direct-superclasses) args)
	(call-next-method))))

(defmethod shared-initialize :around ((instance persistent-object) slot-names &rest initargs &key &allow-other-keys)
  "This seems to be necessary because it is typical for implementations to optimize setting the slots via initforms and initargs in such a way that slot-value-using-class et al aren't used."
  (let* ((class (class-of instance))
	 (persistent-slot-names (persistent-slot-names class)))
    (flet ((persistent-slot-p (item) 
	     (member item persistent-slot-names :test #'eq)))
      (let ((transient-slot-inits 
	     (if (eq slot-names t) ; t means all slots
		 (transient-slot-names class)
		 (remove-if #'persistent-slot-p slot-names)))
	    (persistent-slot-inits
	     (if (eq slot-names t) persistent-slot-names
		 (remove-if-not #'persistent-slot-p slot-names))))
	;; let the implementation initialize the transient slots
	(apply #'call-next-method instance transient-slot-inits initargs)
	;; initialize the persistent slots
	(flet ((initialize-from-initarg (slot-def)
		 (loop for initarg in initargs
		       with slot-initargs = (slot-definition-initargs slot-def)
		       when (member initarg slot-initargs :test #'eq)
		       do 
		       (setf (slot-value-using-class class instance slot-def) 
			     (getf initargs initarg))
		       (return t))))
	  (loop for slot-def in (class-slots class)
		unless (initialize-from-initarg slot-def)
		when (member (slot-definition-name slot-def) persistent-slot-inits :test #'eq)
		unless (slot-boundp-using-class class instance slot-def)
		do
		(let ((initfun (slot-definition-initfunction slot-def)))
		  (when initfun
		    (setf (slot-value-using-class class instance slot-def)
			  (funcall initfun))))))))))

(defmethod slot-value-using-class :around (class (instance persistent-object) (slot-def persistent-slot-definition))
  (declare (ignore class))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-reader instance name)))

(defmethod (setf slot-value-using-class) :around (new-value class (instance persistent-object) (slot-def persistent-slot-definition))
  (declare (ignore class))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-writer new-value instance name)))

(defmethod slot-boundp-using-class :around (class (instance persistent-object) (slot-def persistent-slot-definition))
  (declare (ignore class))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-boundp instance name)))

(defmethod slot-makunbound-using-class :around (class (instance persistent-object) (slot-def persistent-slot-definition))
  (declare (ignore class))
  (buffer-write-int (oid instance) *key-buf*)
  (let ((key-length (serialize (slot-definition-name slot-def) *key-buf*)))
    (db-delete-buffered
     (controller-db *store-controller*) 
     (buffer-stream-buffer *key-buf*)
     key-length
     :transaction *current-transaction*
     :auto-commit *auto-commit*)))
  