;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; controller.lisp -- Lisp interface to a Berkeley DB store
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

(defclass store-controller ()  
  ((path :type (or pathname string)
	 :accessor controller-path
	 :initarg :path)
   (environment :type (or null pointer-void) 
		:accessor controller-environment)
   (db :type (or null pointer-void) :accessor controller-db)
   (root :reader controller-root)
   (instance-cache :accessor instance-cache
		   :initform (make-cache-table :test 'eql)))
  (:documentation "Class of objects responsible for handling
the book-keeping of holding DB handles, the cache, table
creation, counters, locks, the root and garbage collection,
et cetera."))

(defgeneric cache-instance (sc obj))
(defgeneric get-cached-instance (sc oid class-name))
(defgeneric next-oid (sc))
(defgeneric open-controller (sc &key recover recover-fatal thread))
(defgeneric close-controller (sc))

(defun add-to-root (key value &key (store-controller *store-controller*))
  "Add an arbitrary persistent thing to the root, so you can
retrieve it in a later session.  Keys may be arbitrary
persistables as well (though note collection key semantics!)
N.B. this means it (and everything it points to) won't get
gc'd."
  (setf (get-value key (controller-root store-controller)) value))

(defmethod get-from-root (key &key (store-controller *store-controller*))
  "Get a persistent thing from the root."
  (get-value key (controller-root store-controller)))

(defmethod remove-from-root (key &key (store-controller *store-controller*))
  "Get a persistent thing from the root."
  (remove-kv key (controller-root store-controller)))

(defmethod cache-instance ((sc store-controller) obj)
  "Register an instance of a user persistent-class with the
controller."
  (setf (get-cache (oid obj) (instance-cache sc)) obj))

(defmethod get-cached-instance ((sc store-controller) oid class-name)
  (let ((obj (get-cache oid (instance-cache sc))))
    (if obj obj
	;; Should get cached since make-instance calls cache-instance
	(make-instance class-name :from-oid oid))))


;; OID stuff
;; This stuff is all a hack until sequences appear in Sleepycat 4.3
(defvar %oid-entry (uffi:allocate-foreign-object :char 12))
(defvar %oid-lock (uffi:allocate-foreign-object :char 16))

(eval-when (:load-toplevel)
  (loop for c across "%ELEPHANTOID"
	for i from 0 to 11
	do (setf (uffi:deref-array %oid-entry '(:array :char) i)
		 (char-code c)))
  (loop for c across "%ELEPHANTOIDLOCK"
	for i from 0 to 15
	do (setf (uffi:deref-array %oid-lock '(:array :char) i) 
		 (char-code c)))
  )

(defvar %oid-entry-length 12)
(defvar %oid-lock-length 16)

(defmethod next-oid ((sc store-controller))
  (sleepycat::next-counter (controller-environment sc)
			   (controller-db sc)
			   *current-transaction*
			   %oid-entry %oid-entry-length
			   %oid-lock %oid-lock-length))

;; Open/close     
(defmethod open-controller ((sc store-controller) &key (recover nil)
			    (recover-fatal nil) (thread t))
  "Opens the underlying environment and all the necessary
database tables."
  (let ((env (db-env-create)))
    ;; thread stuff?
    (setf (controller-environment sc) env)
    (db-env-open env (controller-path sc) :create t :init-txn t :init-lock t 
		 :init-mpool t :init-log t :thread thread
		 :recover recover :recover-fatal recover-fatal)
    (let ((db (db-create env)))
      (setf (controller-db sc) db)
      (db-open db :file "%ELEPHANT" :database "%ELEPHANTDB" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)
      (let ((root (make-instance 'btree :from-oid -1)))
	(setf (slot-value sc 'root) root)
	(let ((*auto-commit* t))
	  (unless (db-get-key-buffered db %oid-entry %oid-entry-length)
	    (buffer-write-int 0 *out-buf*)
	    (db-put-buffered db %oid-entry %oid-entry-length 
			     (buffer-stream-buffer *out-buf*) 4
			     :auto-commit t)
	    (finish-buffer *out-buf*)))
	sc))))

(defmethod close-controller ((sc store-controller))
  "Close the db handles and environment.  Tries to wipe out
references to the db handles."
  (when (slot-value sc 'root)
    ;; no root
    (setf (slot-value sc 'root) nil)
    ;; clean instance cache
    (setf (instance-cache sc) (make-cache-table :test 'eql))
    ;; close environment
    (db-close (controller-db sc))
    (setf (controller-db sc) nil)
    (db-env-close (controller-environment sc))
    (setf (controller-environment sc) nil)
    nil))

(defmacro with-open-controller ((&optional (sc '*store-controller*))
				&body body)
  `(unwind-protect
       (progn
	 (let (*store-controller* (open-controller ,sc))
	   (declare (special *store-controller*))
	   ,@body))
     (close-controller ,sc)))

(defun open-store (path)
  (setq *store-controller* (make-instance 'store-controller :path path))
  (open-controller *store-controller*))

(defun close-store ()
  (close-controller *store-controller*))

(defmacro with-open-store ((path) &body body)
  (let ((sc (gensym)))
    `(let ((,sc (make-instance 'store-controller :path ,path)))
      (unwind-protect
	   (progn
	     (let ((*store-controller* ,sc))
	       (declare (special *store-controller*))
	       (open-controller *store-controller*)
	       ,@body))
	(close-controller ,sc)))))

