/*
;;;
;;; libsleepycat.c -- C wrappers for Sleepycat for FFI
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;
8?

/* Pointer arithmetic utility functions */

int read_int(char *buf, int offset) {
  return *(int*)(buf + offset);
}

unsigned int read_uint(char *buf, int offset) {
  return *(unsigned int*)(buf + offset);
}

float read_float(char *buf, int offset) {
  return *(float*)(buf + offset);
}

double read_double(char *buf, int offset) {
  return *(double*)(buf + offset);
}

void write_int(char *buf, int num, int offset) {
  *(int*)(buf + offset) = num;
}

void write_uint(char *buf, unsigned int num, int offset) {
  *(unsigned int*)(buf + offset) = num;
}

void write_float(char *buf, float num, int offset) {
  *(float*)(buf + offset) = num;
}

void write_double(char *buf, double num, int offset) {
  *(double*)(buf + offset) = num;
}

char *offset_charp(char *p, int offset) {
  return p + offset;
}

#include <string.h>

void copy_buf(char *dest, int dest_offset, char *src, int src_offset, 
	      int length) {
  memcpy(dest + dest_offset, src + src_offset, length);
}


/* Sleepycat stuff */

#include <db.h>

/* Environment */

/* All "creation" functions return the create object, not the errno.
   This simplifies FFI handling. */

/* These next two functions are also needed because in db42 these
   are #define macros */
DB_ENV *db_env_cr(u_int32_t flags, int *errno) {
  DB_ENV *envp;
  *errno = db_env_create(&envp, flags);
  return envp; 
}

char * db_strerr(int error) {
  return db_strerror(error);
}

int db_env_close(DB_ENV *env, u_int32_t flags) {
  return env->close(env, flags);
}

int db_env_open(DB_ENV *env, char *home, u_int32_t flags, int mode) {
  return env->open(env, home, flags, mode);
}

int db_env_dbremove(DB_ENV *env, DB_TXN *txnid, char *file, char *database,
		    u_int32_t flags) {
  return env->dbremove(env, txnid, file, database, flags);
}

int db_env_dbrename(DB_ENV *env, DB_TXN *txnid, char *file, char *database,
		    char *newname, u_int32_t flags) {
  return env->dbrename(env, txnid, file, database, newname, flags);
}

int db_env_remove(DB_ENV *env, char *home, u_int32_t flags) {
  return env->remove(env, home, flags);
}

int db_env_set_flags(DB_ENV *dbenv, u_int32_t flags, int onoff) {
  return dbenv->set_flags(dbenv, flags, onoff);
}

int db_env_get_flags(DB_ENV *dbenv, u_int32_t *flagsp) {
  return dbenv->get_flags(dbenv, flagsp);
}

/* Database */

DB *db_cr(DB_ENV *dbenv, u_int32_t flags, int *errno) {
  DB *dbp;
  *errno = db_create(&dbp, dbenv, flags);
  return dbp;
}

int db_close(DB *db, u_int32_t flags) {
  return db->close(db, flags);
}

int db_open(DB *db, DB_TXN *txnid, char *file, char *database, DBTYPE type,
	    u_int32_t flags, int mode) {
  return db->open(db, txnid, file, database, type, flags, mode);
}

int db_remove(DB *db, char *file, char *database, u_int32_t flags) {
  return db->remove(db, file, database, flags);
}

int db_rename(DB *db, char *file, char *database, char *newname,
	      u_int32_t flags) {
  return db->rename(db, file, database, newname, flags);
}

int db_sync(DB *db, u_int32_t flags) {
  return db->sync(db, flags);
}

int db_truncate(DB *db, DB_TXN *txnid, u_int32_t *countp, u_int32_t flags) {
  return db->truncate(db, txnid, countp, flags);
}

int db_set_pagesize(DB *db, u_int32_t pagesize) {
  return db->set_pagesize(db, pagesize);
}

int db_get_pagesize(DB *db, u_int32_t *pagesizep) {
  return db->get_pagesize(db, pagesizep);
}

/* Accessors */
/* We manage our own buffers (DB_DBT_USERMEM). */

int db_get_raw(DB *db, DB_TXN *txnid, 
	       char *key, u_int32_t key_length,
	       char *buffer, u_int32_t buffer_length,
	       u_int32_t flags, u_int32_t *result_length) {
  DBT DBTKey, DBTDatum;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTDatum, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_length;
  DBTDatum.data = buffer;
  DBTDatum.ulen = buffer_length;
  DBTDatum.flags |= DB_DBT_USERMEM;
  
  ret = db->get(db, txnid, &DBTKey, &DBTDatum, flags);
  *result_length = DBTDatum.size;
  
  return ret;
}

int db_put_raw(DB *db, DB_TXN *txnid, 
	       char *key, u_int32_t key_length,
	       char *datum, u_int32_t datum_length,
	       u_int32_t flags) {
  DBT DBTKey, DBTDatum;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTDatum, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_length;
  DBTDatum.data = datum;
  DBTDatum.size = datum_length;
  
  return db->put(db, txnid, &DBTKey, &DBTDatum, flags);
}

int db_del(DB *db, DB_TXN *txnid, 
	   char *key, u_int32_t key_length,
	   u_int32_t flags) {
  DBT DBTKey;
  
  memset(&DBTKey, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_length;
  return db->del(db, txnid, &DBTKey, flags);
}


/* Cursors */

DBC * db_cursor(DB *db, DB_TXN *txnid, u_int32_t flags, int *errno) {
  DBC *cursor;
  *errno = db->cursor(db, txnid, &cursor, flags);
  return cursor;
}

int db_cursor_close(DBC *cursor) {
  return cursor->c_close(cursor);
}

int db_cursor_del(DBC *cursor, u_int32_t flags) {
  return cursor->c_del(cursor, flags);
}

DBC * db_cursor_dup(DBC *cursor, u_int32_t flags, int *errno) {
  DBC *dup;
  *errno = cursor->c_dup(cursor, &dup, flags);
  return dup;
}

int db_cursor_get_raw(DBC *cursor, 
		      char *keybuf, u_int32_t keybuf_length,
		      char *buffer, u_int32_t buffer_length,
		      u_int32_t flags, u_int32_t *key_length,
		      u_int32_t *result_length) {
  DBT DBTKey, DBTDatum;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTDatum, 0, sizeof(DBT));
  DBTKey.data = keybuf;
  DBTKey.ulen = keybuf_length;
  DBTKey.flags |= DB_DBT_USERMEM;
  DBTDatum.data = buffer;
  DBTDatum.ulen = buffer_length;
  DBTDatum.flags |= DB_DBT_USERMEM;
  
  ret = cursor->c_get(cursor, &DBTKey, &DBTDatum, flags);
  *key_length = DBTKey.size;
  *result_length = DBTDatum.size;
  
  return ret;
}

int db_cursor_put_raw(DBC *cursor,
		      char *key, u_int32_t key_length,
		      char *datum, u_int32_t datum_length,
		      u_int32_t flags) {
  DBT DBTKey, DBTDatum;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTDatum, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_length;
  DBTDatum.data = datum;
  DBTDatum.size = datum_length;
  
  return cursor->c_put(cursor, &DBTKey, &DBTDatum, flags);
}


/* Bulk retrieval */

int db_cursor_get_multiple_key(DBC *cursor, 
			       char *keybuf, u_int32_t keybuf_length,
			       char *buffer, u_int32_t buffer_length,
			       u_int32_t flags, u_int32_t *key_length,
			       u_int32_t *result_length,
			       void **pointer, DBT **data) {
  DBT DBTKey, DBTDatum;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTDatum, 0, sizeof(DBT));
  DBTKey.data = keybuf;
  DBTKey.ulen = keybuf_length;
  DBTKey.flags |= DB_DBT_USERMEM;
  DBTDatum.data = buffer;
  DBTDatum.ulen = buffer_length;
  DBTDatum.flags |= DB_DBT_USERMEM;
  
  flags |= DB_MULTIPLE_KEY;
  ret = cursor->c_get(cursor, &DBTKey, &DBTDatum, flags);
  *key_length = DBTKey.size;
  *result_length = DBTDatum.size;
  if ((DBTKey.size <= DBTKey.ulen) && (DBTDatum.size <= DBTDatum.ulen)) {
    **data = DBTDatum;
    DB_MULTIPLE_INIT(*pointer, *data);
  }
				    
  return ret;
}

void db_multiple_key_next(void *pointer, DBT *data,
			  char **key, u_int32_t *key_length,
			  char **result, u_int32_t *result_length) {
  DB_MULTIPLE_KEY_NEXT(pointer, data, 
		       *key, *key_length, 
		       *result, *result_length);
}

/* Transactions */

DB_TXN * db_txn_begin(DB_ENV *env, DB_TXN *parent, 
		      u_int32_t flags, int *errno) {
  DB_TXN * p;
  *errno = env->txn_begin(env, parent, &p, flags);
  return p;
}

int db_txn_abort(DB_TXN *txnid) {
  return txnid->abort(txnid);
}

int db_txn_commit(DB_TXN *txnid, u_int32_t flags) {
  return txnid->commit(txnid, flags);
}


int db_txnp_begin(DB_ENV *env, DB_TXN *parent, DB_TXN **txnp,
		 u_int32_t flags) {
  return env->txn_begin(env, parent, txnp, flags);
}

/* Locks and timeouts */

u_int32_t db_txn_id(DB_TXN *tid) {
  return tid->id(tid);
}

int db_env_lock_id(DB_ENV *env, u_int32_t *idp) {
  return env->lock_id(env, idp);
}

int db_env_lock_id_free(DB_ENV *env, u_int32_t id) {
  return env->lock_id_free(env, id);
}

int db_env_lock_get(DB_ENV *env, u_int32_t locker,
		    u_int32_t flags, char *object, u_int32_t object_length,
		    const db_lockmode_t lock_mode, DB_LOCK *lock) {
  DBT DBTObject;
  memset(&DBTObject, 0, sizeof(DBT));
  DBTObject.data = object;
  DBTObject.size = object_length;

  return env->lock_get(env, locker, flags, &DBTObject, lock_mode, lock);
}

int db_env_lock_put(DB_ENV *env, DB_LOCK *lock) {
  return env->lock_put(env, lock);
}

int db_env_lock_vec(DB_ENV *env, u_int32_t locker, u_int32_t flags,
		    DB_LOCKREQ list[], int nlist, DB_LOCKREQ **elistp) {
  return env->lock_vec(env, locker, flags, list, nlist, elistp);
}

/* db_timeout_t = u_int32_t */
int db_env_set_timeout(DB_ENV *env, db_timeout_t timeout, u_int32_t flags) {
  return env->set_timeout(env, timeout, flags);
}

int db_env_get_timeout(DB_ENV *env, db_timeout_t *timeoutp, u_int32_t flags) {
  return env->get_timeout(env, timeoutp, flags);
}

int db_env_set_lk_detect(DB_ENV *env, u_int32_t detect) {
    return env->set_lk_detect(env, detect);
}

int db_env_get_lk_detect(DB_ENV *env, u_int32_t *detectp) {
    return env->get_lk_detect(env, detectp);
}

int db_env_lock_detect(DB_ENV *env, u_int32_t flags, u_int32_t atype,
		   int *aborted) {
    return env->lock_detect(env, flags, atype, aborted);
}

/* Poor man's counters */

int next_counter(DB_ENV *env, DB *db, DB_TXN *parent,
		 char *key, u_int32_t key_length,
		 char *lockid, u_int32_t lockid_length) {
  DB_LOCK lock;
  DBT DBTKey, DBTData; 
  DB_TXN *tid; 
  int counter, tries, ret, t_ret, lockheld;
  u_int32_t id;

  /* Initialization. */ 
  memset(&lock, 0, sizeof(lock)); 
  memset(&DBTKey, 0, sizeof(DBTKey)); 
  memset(&DBTData, 0, sizeof(DBTData)); 
  DBTKey.data = key;
  DBTKey.size = key_length;
  DBTData.data = lockid;
  DBTData.size = lockid_length;

  tries = 0;

 loop:
  lockheld = 0;

  /* Begin the transaction. */ 
  if ((ret = env->txn_begin(env, parent, &tid, 0)) != 0) { 
    env->err(env, ret, "DB_ENV->txn_begin"); 
    return (-1); 
  }

  id = tid->id(tid);

  if ((ret = env->lock_get(env, id, 0, &DBTData, DB_LOCK_WRITE, &lock)) != 0)
    goto fail;

  lockheld = 1;

  memset(&DBTData, 0, sizeof(DBTData)); 
  DBTData.data = &counter;
  DBTData.ulen = sizeof(counter);
  DBTData.flags |= DB_DBT_USERMEM;

  if ((ret = db->get(db, tid, &DBTKey, &DBTData, 0)) != 0) 
    goto fail;

  ++counter;

  memset(&DBTData, 0, sizeof(DBTData)); 
  DBTData.data = &counter;
  DBTData.size = sizeof(counter);

  if ((ret = db->put(db, tid, &DBTKey, &DBTData, 0)) != 0) 
    goto fail;

  if ((ret = env->lock_put(env, &lock)) != 0) 
    goto fail;

  if ((ret = tid->commit(tid, DB_TXN_NOSYNC)) != 0) { 
    env->err(env, ret, "DB_TXN->commit"); 
    return (-2); 
  } 
  return (counter); 


 fail:
  if (lockheld)
    if ((ret = env->lock_put(env, &lock)) != 0) 
      return (-3);

  /* Abort and retry the operation. */ 
  if ((t_ret = tid->abort(tid)) != 0) { 
    env->err(env, t_ret, "DB_TXN->abort"); 
    return (-4);
  } 
  if (tries++ == 100)
    return (-5); 
  goto loop;
}
