;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; metaclasses.lisp -- persistent objects via metaobjects
;;; 
;;; Initial version 8/26/2004 by Andrew Blumberg
;;; <ablumberg@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

(defclass persistent ()
  ((%oid :accessor oid :initarg :from-oid))
  (:documentation 
   "Abstract superclass for all persistent classes (common
to user-defined classes and collections.)"))

(defclass persistent-metaclass (standard-class)
  ())

(defclass persistent-slot-definition (standard-slot-definition)
  ())

(defclass persistent-direct-slot-definition (standard-direct-slot-definition persistent-slot-definition)
  ())

(defclass persistent-effective-slot-definition (standard-effective-slot-definition persistent-slot-definition)
  ())

(defclass transient-slot-definition (standard-slot-definition)
  ((transient :initform t :initarg :transient :allocation :class)))

(defclass transient-direct-slot-definition (standard-direct-slot-definition transient-slot-definition)
  ())

(defclass transient-effective-slot-definition (standard-effective-slot-definition transient-slot-definition)
  ())

(defgeneric transient (slot))

(defmethod transient ((slot standard-direct-slot-definition))
  t)

(defmethod transient ((slot persistent-direct-slot-definition))
  nil)

(defmethod slot-definition-allocation ((slot-definition persistent-slot-definition))
  :class)

(defmethod direct-slot-definition-class ((class persistent-metaclass) &rest initargs)
  (let ((allocation-key (getf initargs :allocation))
	(transient-p (getf initargs :transient)))
    (when (consp transient-p) (setq transient-p (car transient-p)))
    (cond ((and (eq allocation-key :class) transient-p)
	   (find-class 'transient-direct-slot-definition))
	  ((and (eq allocation-key :class) (not transient-p))
	   (error "Persistent class slots are not supported, try :transient t."))
	  (transient-p
	   (find-class 'transient-direct-slot-definition))
	  (t
	   (find-class 'persistent-direct-slot-definition)))))

(defmethod validate-superclass ((class persistent-metaclass) (super standard-class))
  t)

(defmethod validate-superclass ((class standard-class) (super persistent-metaclass))
  nil)

(defgeneric persistent-p (class))

(defmethod persistent-p ((class t))
  nil)

(defmethod persistent-p ((class persistent-metaclass))
  t)

(defmethod persistent-p ((class persistent-slot-definition))
  t)

(defmethod effective-slot-definition-class ((class persistent-metaclass) &rest initargs)
  (let ((transient-p (getf initargs :transient)))
    (when (consp transient-p) (setq transient-p (car transient-p)))
    (cond (transient-p
	   (find-class 'transient-effective-slot-definition))
	  (t
	   (find-class 'persistent-effective-slot-definition)))))

#+(or cmu sbcl) 
(defgeneric ensure-storage-exists (class slot-definition slot-name))

#+(or cmu sbcl) 
(defmethod ensure-storage-exists (class slot-definition slot-name)
  nil)

#+(or cmu sbcl) 
(defmethod ensure-storage-exists (class (slot-definition persistent-slot-definition) slot-name)
  (let ((use-class (or (slot-definition-allocation-class slot-definition)
		       class)))
    (when (not (assoc slot-name (class-slot-cells use-class)))
      (setf (plist-value use-class 'class-slot-cells) 
	    (append
	     (plist-value use-class 'class-slot-cells)
	     (list (cons slot-name +slot-unbound+)))))))

#+(or cmu sbcl) 
(defmethod compute-effective-slot-definition :around ((class persistent-metaclass) slot-name direct-slot-definitions)
  (let ((slot-definition (call-next-method)))
    (ensure-storage-exists class slot-definition slot-name)
    slot-definition))

#+openmcl
(defmethod compute-effective-slot-definition ((class persistent-metaclass) slot-name direct-slot-definitions)
  (declare (ignore slot-name))
  (apply #'make-effective-slot-definition class
	 (compute-effective-slot-definition-initargs 
	  class direct-slot-definitions)))

#+openmcl
(defmethod compute-effective-slot-definition-initargs ((class slots-class)
						       direct-slots)
  (let* ((name (loop for s in direct-slots
		     when s
		     do (return (slot-definition-name s))))
	 (initer (dolist (s direct-slots)
                   (when (%slot-definition-initfunction s)
                     (return s))))
         (documentor (dolist (s direct-slots)
                       (when (%slot-definition-documentation s)
                         (return s))))
         (first (car direct-slots))
         (initargs (let* ((initargs nil))
                     (dolist (dslot direct-slots initargs)
                       (dolist (dslot-arg (%slot-definition-initargs  dslot))
                         (pushnew dslot-arg initargs :test #'eq))))))
    (list
     :name name
     :allocation (%slot-definition-allocation first)
     :documentation (when documentor (nth-value
                                      1
                                      (%slot-definition-documentation
                                       documentor)))
     :class (%slot-definition-class first)
     :initargs initargs
     :initfunction (if initer (%slot-definition-initfunction initer))
     :initform (if initer (%slot-definition-initform initer))
     :type (or (%slot-definition-type first) t))))

(defun ensure-transient-chain (slot-definitions initargs)
  (declare (ignore initargs))
  (loop for slot-definition in slot-definitions
	always (transient slot-definition)))

(defmethod compute-effective-slot-definition-initargs ((class persistent-metaclass) slot-definitions)
  (let ((initargs (call-next-method)))
    (if (ensure-transient-chain slot-definitions initargs)
	(append initargs '(:transient t))
	(progn
	  (setf (getf initargs :allocation) :class)
	  initargs))))

(defmacro persistent-slot-reader (instance name)
  `(progn
    (buffer-write-int (oid ,instance) *key-buf*)
    (let* ((key-length (serialize ,name *key-buf*))
	   (buf (db-get-key-buffered 
		 (controller-db *store-controller*) 
		 (buffer-stream-buffer *key-buf*)
		 key-length)))
      (if buf (deserialize buf)
	  #+cmu
	  (error 'unbound-slot :instance ,instance :slot ,name)
	  #-cmu
	  (error 'unbound-slot :instance ,instance :name ,name)))))

#+(or cmu sbcl)
(defun make-persistent-reader (name)
  (lambda (instance)
    (declare (optimize (speed 3))
	     (type persistent-object instance))
    (persistent-slot-reader instance name)))

(defmacro persistent-slot-writer (new-value instance name)
  `(progn
    (buffer-write-int (oid ,instance) *key-buf*)
    (let ((key-length (serialize ,name *key-buf*))
	  (val-length (serialize ,new-value *out-buf*)))
      (db-put-buffered (controller-db *store-controller*) 
		       (buffer-stream-buffer *key-buf*) key-length
		       (buffer-stream-buffer *out-buf*) val-length
		       :transaction *current-transaction*
		       :auto-commit *auto-commit*)
      ,new-value)))

#+(or cmu sbcl)
(defun make-persistent-writer (name)
  (lambda (new-value instance)
    (declare (optimize (speed 3))
	     (type persistent-object instance))
    (persistent-slot-writer new-value instance name)))

(defmacro persistent-slot-boundp (instance name)
  `(progn
    (buffer-write-int (oid ,instance) *key-buf*)
    (let* ((key-length (serialize ,name *key-buf*))
	   (buf (db-get-key-buffered 
		 (controller-db *store-controller*) 
		 (buffer-stream-buffer *key-buf*)
		 key-length)))
      (if buf T nil))))

#+(or cmu sbcl)
(defun make-persistent-slot-boundp (name)
  (lambda (instance)
    (declare (optimize (speed 3))
	     (type persistent-object instance))
    (persistent-slot-boundp instance name)))

#+(or cmu sbcl)
(defmethod initialize-internal-slot-functions ((slot-def persistent-slot-definition))
  (let ((name (slot-definition-name slot-def)))
    (setf (slot-definition-reader-function slot-def)
	  (make-persistent-reader name))
    (setf (slot-definition-writer-function slot-def)
	  (make-persistent-writer name))
    (setf (slot-definition-boundp-function slot-def)
	  (make-persistent-slot-boundp name)))
  slot-def)

(defun persistent-slot-names (class)
  (let ((slot-definitions (class-slots class)))
    (loop for slot-definition in slot-definitions
	  when (equalp (class-of slot-definition) (find-class 'persistent-effective-slot-definition))
	  collect (slot-definition-name slot-definition))))

(defun transient-slot-names (class)
  (let ((slot-definitions (class-slots class)))
    (loop for slot-definition in slot-definitions
	  unless (persistent-p slot-definition)
	  collect (slot-definition-name slot-definition))))