;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; serializer.lisp -- convert Lisp data to/from byte arrays
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

(declaim (inline int-byte-spec
		 ;serialize deserialize
		 slots-and-values
		 deserialize-bignum))

(def-type foreign-char :char)

;; Constants

(defconstant +fixnum+                (char-code #\f))
(defconstant +nil+                   (char-code #\N))
(defconstant +symbol+                (char-code #\S))
(defconstant +string+                (char-code #\s))
(defconstant +persistent+            (char-code #\P))
(defconstant +single-float+          (char-code #\F))
(defconstant +double-float+          (char-code #\D))
(defconstant +char+                  (char-code #\c))
(defconstant +pathname+              (char-code #\p))
(defconstant +positive-bignum+       (char-code #\B))
(defconstant +negative-bignum+       (char-code #\b))
(defconstant +rational+              (char-code #\r))
(defconstant +cons+                  (char-code #\C))
(defconstant +hash-table+            (char-code #\H))
(defconstant +object+                (char-code #\O))

(defconstant +array+                 (char-code #\A))

(defconstant +fill-pointer-p+ #x40)
(defconstant +adjustable-p+ #x80)


(defun serialize (frob bs)
  (declare (optimize (speed 3) (safety 0)))
  (setq *lisp-obj-id* 0)
  (clrhash *circularity-hash*)
  (labels 
      ((%serialize (frob)
	 (declare (optimize (speed 3) (safety 0)))
	 (etypecase frob
	   (fixnum
	    (buffer-write-byte +fixnum+ bs)
	    (buffer-write-int frob bs))
	   (null
	    (buffer-write-byte +nil+ bs))
	   (symbol
	    (let ((s (symbol-name frob)))
	      (declare (type string s) (dynamic-extent s))
	      (buffer-write-byte +symbol+ bs)
	      (buffer-write-int (byte-length s) bs)
	      (buffer-write-string s bs)
	      (let ((package (symbol-package frob)))
		(if package
		    (%serialize (package-name package))
		    (%serialize nil)))))
	   (string
	    (buffer-write-byte +string+ bs)
	    (buffer-write-int (byte-length frob) bs)
	    (buffer-write-string frob bs))
	   (persistent
	    (buffer-write-byte +persistent+ bs)
	    (buffer-write-int (oid frob) bs)
	    (%serialize (type-of frob)))
	   #-(and :lispworks (or :win32 :linux))
	   (single-float
	    (buffer-write-byte +single-float+ bs)
	    (buffer-write-float frob bs))
	   (double-float
	    (buffer-write-byte +double-float+ bs)
	    (buffer-write-double frob bs))
	   (character
	    (buffer-write-byte +char+ bs)
	    ;; might be wide!
	    (buffer-write-uint (char-code frob) bs))
	   (pathname
	    (let ((s (namestring frob)))
	      (declare (type string s) (dynamic-extent s))
	      (buffer-write-byte +pathname+ bs)
	      (buffer-write-int (byte-length s) bs)
	      (buffer-write-string s bs)))
	   (integer
	    (let* ((num (abs frob))
		   (word-size (ceiling (/ (integer-length num) 32)))
		   (needed (* word-size 4)))
	      (declare (type fixnum word-size needed))
	      (if (< frob 0) 
		  (buffer-write-byte +negative-bignum+ bs)
		  (buffer-write-byte +positive-bignum+ bs))
	      (buffer-write-int needed bs)
	      (loop for i fixnum from 0 below word-size 
		    ;; this ldb is consing on CMUCL!
		    ;; there is an OpenMCL function which should work 
		    ;; and non-cons
		    do
		    #+(or cmu sbcl)
		    (buffer-write-uint (%bignum-ref num i) bs)
		    #+(or allegro lispworks openmcl)
		    (buffer-write-uint (ldb (int-byte-spec i) num) bs))))
	   (rational
	    (buffer-write-byte +rational+ bs)
	    (%serialize (numerator frob))
	    (%serialize (denominator frob)))
	   (cons
	    (buffer-write-byte +cons+ bs)
	    (let ((idp (gethash frob *circularity-hash*)))
	      (if idp (buffer-write-int idp bs)
		  (progn
		    (buffer-write-int (incf *lisp-obj-id*) bs)
		    (setf (gethash frob *circularity-hash*) *lisp-obj-id*)
		    (%serialize (car frob))
		    (%serialize (cdr frob))))))
	   (hash-table
	    (buffer-write-byte +hash-table+ bs)
	    (let ((idp (gethash frob *circularity-hash*)))
	      (if idp (buffer-write-int idp bs)
		  (progn
		    (buffer-write-int (incf *lisp-obj-id*) bs)
		    (setf (gethash frob *circularity-hash*) *lisp-obj-id*)
		    (%serialize (hash-table-test frob))
		    (%serialize (hash-table-rehash-size frob))
		    (%serialize (hash-table-rehash-threshold frob))
		    (%serialize (hash-table-count frob))
		    (loop for key being the hash-key of frob
			  using (hash-value value)
			  do 
			  (%serialize key)
			  (%serialize value))))))
	   (standard-object
	    (buffer-write-byte +object+ bs)
	    (let ((idp (gethash frob *circularity-hash*)))
	      (if idp (buffer-write-int idp bs)
		  (progn
		    (buffer-write-int (incf *lisp-obj-id*) bs)
		    (setf (gethash frob *circularity-hash*) *lisp-obj-id*)
		    (%serialize (type-of frob))
		    (let ((svs (slots-and-values frob)))
		      (declare (dynamic-extent svs))
		      (%serialize (/ (length svs) 2))
		      (loop for item in svs
			    do (%serialize item)))))))
	   (array
	    (buffer-write-byte +array+ bs)
	    (let ((idp (gethash frob *circularity-hash*)))
	      (if idp (buffer-write-int idp bs)
		  (progn
		    (buffer-write-int (incf *lisp-obj-id*) bs)
		    (setf (gethash frob *circularity-hash*) *lisp-obj-id*)
		    (buffer-write-byte 
		     (logior (byte-from-array-type (array-element-type frob))
			     (if (array-has-fill-pointer-p frob) 
				 +fill-pointer-p+ 0)
			     (if (adjustable-array-p frob) 
				 +adjustable-p+ 0))
		     bs)
		    (let ((rank (array-rank frob)))
		      (buffer-write-int rank bs)
		      (loop for i fixnum from 0 below rank
			    do (buffer-write-int (array-dimension frob i) 
						 bs)))
		    (when (array-has-fill-pointer-p frob)
		      (buffer-write-int (fill-pointer frob) bs))
		    (loop for i fixnum from 0 below (array-total-size frob)
			  do
			  (%serialize (row-major-aref frob i)))))))
	   )))
    (%serialize frob)
    (finish-buffer bs)))

(defun slots-and-values (o)
  (declare (optimize (speed 3) (safety 0)))
  (loop for sd in (compute-slots (class-of o))
	for slot-name = (slot-definition-name sd)
	with ret = ()
	do
	(when (and (slot-boundp o slot-name)
		   (eq :instance
		       (slot-definition-allocation sd)))
	  (push (slot-value o slot-name) ret)
	  (push slot-name ret))
	finally (return ret)))

(defun deserialize (buf)
  (declare (optimize (speed 3) (safety 0))
	   (type (or null array-or-pointer-char) buf))
  (unless buf (return-from deserialize nil))
  (setf (buffer-stream-buffer *in-buf*) buf)
  (setf (buffer-stream-position *in-buf*) 0)
  (setq *lisp-obj-id* 0)
  (clrhash *circularity-hash*)
  (labels 
      ((%deserialize (bs)
	 (declare (optimize (speed 3) (safety 0))
		  (type buffer-stream bs))
	 (let ((tag (buffer-read-byte bs)))
	   (declare (type foreign-char tag))
	   (cond
	     ((= tag +fixnum+) 
	      (buffer-read-fixnum bs))
	     ((= tag +nil+) nil)
	     ((= tag +symbol+)
	      (let ((name (buffer-read-string bs (buffer-read-fixnum bs)))
		    (maybe-package-name (%deserialize bs)))
		(if maybe-package-name
		    (intern name (find-package maybe-package-name))
		    (make-symbol name))))
	     ((= tag +string+)
	      (buffer-read-string bs (buffer-read-fixnum bs)))
	     ((= tag +persistent+)
	      (get-cached-instance *store-controller*
				   (buffer-read-fixnum bs)
				   (%deserialize bs)))
	     ((= tag +single-float+) 
	      (buffer-read-float bs))
	     ((= tag +double-float+) 
	      (buffer-read-double bs))
	     ((= tag +char+)
	      (code-char (buffer-read-uint bs)))
	     ((= tag +pathname+)
	      (parse-namestring 
	       (or (buffer-read-string bs (buffer-read-fixnum bs)) "")))
	     ((= tag +positive-bignum+) 
	      (deserialize-bignum bs (buffer-read-fixnum bs) t))
	     ((= tag +negative-bignum+) 
	      (deserialize-bignum bs (buffer-read-fixnum bs) nil))
	     ((= tag +rational+) 
	      (/ (the integer (%deserialize bs)) 
		 (the integer (%deserialize bs))))
	     ((= tag +cons+)
	      (let* ((id (buffer-read-fixnum bs))
		     (maybe-cons (gethash id *circularity-hash*)))
		(if maybe-cons maybe-cons
		    (let ((c (cons nil nil)))
		      (setf (gethash id *circularity-hash*) c)
		      (setf (car c) (%deserialize bs))
		      (setf (cdr c) (%deserialize bs))
		      c))))
	     ((= tag +hash-table+)
	      (let* ((id (buffer-read-fixnum bs))
		     (maybe-hash (gethash id *circularity-hash*)))
		(if maybe-hash maybe-hash
		    (let ((h (make-hash-table :test (%deserialize bs)
					      :rehash-size (%deserialize bs)
					      :rehash-threshold 
					      (%deserialize bs))))
		      (setf (gethash id *circularity-hash*) h)
		      (loop for i fixnum from 0 below (%deserialize bs)
			    do
			    (setf (gethash (%deserialize bs) h) 
				  (%deserialize bs)))
		      h))))
	     ((= tag +object+)
	      (let* ((id (buffer-read-fixnum bs))
		     (maybe-o (gethash id *circularity-hash*)))
		(if maybe-o maybe-o
		    (let ((o (make-instance (%deserialize bs))))
		      (setf (gethash id *circularity-hash*) o)
		      (loop for i fixnum from 0 below (%deserialize bs)
			    do
			    (setf (slot-value o (%deserialize bs))
				  (%deserialize bs)))
		      o))))
	     ((= tag +array+)
	      (let* ((id (buffer-read-fixnum bs))
		     (maybe-array (gethash id *circularity-hash*)))
		(if maybe-array maybe-array
		    (let* ((flags (buffer-read-byte bs))
			   (a (make-array 
			       (loop for i fixnum from 0 below 
				     (buffer-read-int bs)
				     collect (buffer-read-int bs))
			       :element-type (array-type-from-byte 
					      (logand #x3f flags))
			       :fill-pointer (/= 0 (logand +fill-pointer-p+ 
							   flags))
			       :adjustable (/= 0 (logand +adjustable-p+ 
							 flags)))))
		      (when (array-has-fill-pointer-p a)
			(setf (fill-pointer a) (buffer-read-int bs)))
		      (setf (gethash id *circularity-hash*) a)
		      (loop for i fixnum from 0 below (array-total-size a)
			    do
			    (setf (row-major-aref a i) (%deserialize bs)))
		      a))))		    
	     (t (error "deserialize fubar!"))))))
    (%deserialize *in-buf*)))

(defun deserialize-bignum (bs length positive)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type fixnum length)
	   (type boolean positive))
  (loop for i from 0 below (/ length 4)
	for byte-spec = (int-byte-spec i)
	with num integer = 0 
	do
	(setq num (dpb (buffer-read-uint bs) byte-spec num))
	finally (return (if positive num (- num)))))


;; array type tags

(declaim (type hash-table array-type-to-byte byte-to-array-type))
(defvar array-type-to-byte (make-hash-table :test 'equalp))
(defvar byte-to-array-type (make-hash-table :test 'equalp))

(setf (gethash 'T array-type-to-byte) #x00)
(setf (gethash 'base-char array-type-to-byte) #x01)
(setf (gethash 'character array-type-to-byte) #x02)
(setf (gethash 'single-float array-type-to-byte) #x03)
(setf (gethash 'double-float array-type-to-byte) #x04)
(setf (gethash '(complex single-float) array-type-to-byte) #x05)
(setf (gethash '(complex double-float) array-type-to-byte) #x06)
(setf (gethash 'fixnum array-type-to-byte) #x07)
(setf (gethash 'bit array-type-to-byte) #x08)

(defun type= (t1 t2)
  (and (subtypep t1 t2) (subtypep t2 t1)))

(let ((counter 8))
  (loop for i from 2 to 65
	for spec = (list 'unsigned-byte i)
	for uspec = (upgraded-array-element-type spec)
	when (type= spec uspec)
	do
	(setf (gethash spec array-type-to-byte) (incf counter)))
  (loop for i from 2 to 65
	for spec = (list 'signed-byte i)
	for uspec = (upgraded-array-element-type spec)
	when (type= spec uspec)
	do
	(setf (gethash spec array-type-to-byte) (incf counter))))

(loop for key being the hash-key of array-type-to-byte 
      using (hash-value value)
      do
      (setf (gethash value byte-to-array-type) key))

(defun array-type-from-byte (b)
  (gethash b byte-to-array-type))

(defun byte-from-array-type (ty)
  (the (unsigned-byte 8) (gethash ty array-type-to-byte)))

(defun int-byte-spec (position)
  (declare (optimize (speed 3) (safety 0))
	   (type (unsigned-byte 24) position))
  #+(or cmu scl sbcl allegro)
  (progn (setf (cdr *resourced-byte-spec*) (* 32 position)) 
	 *resourced-byte-spec*)
  #-(or cmu scl sbcl allegro)
  (byte 32 (* 32 position))
  )
