;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; utils.lisp -- utility functions
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(in-package "ELEPHANT")
(eval-when (:compile-toplevel :load-toplevel :execute)
  (use-package "UFFI"))

(declaim (inline ;resize-buffer-stream 
		 finish-buffer
		 buffer-write-byte buffer-write-int buffer-write-uint
		 buffer-write-float buffer-write-double buffer-write-string
		 buffer-read-byte buffer-read-fixnum buffer-read-int
		 buffer-read-uint buffer-read-float buffer-read-double 
		 buffer-read-string)
	 (type fixnum *lisp-obj-id*)
	 (type hash-table *circularity-hash*)
	 (type boolean *auto-commit*))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; buffer-streams
;;;
;;; a stream-like interface for our buffers; methods are
;;; below.  ultimately we might want a gray / simple -stream
;;; for real, for now who cares?

(defstruct buffer-stream
  (buffer (allocate-foreign-object :char 1) :type array-or-pointer-char)
  (length 0 :type fixnum)
  (position 0 :type fixnum))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Thread-local specials

(defparameter *store-controller* nil 
  "The store controller which persistent objects talk to.")

;; Specials which control persistent objects
(defvar *auto-commit* T)

(declaim (type buffer-stream *out-buf* *key-buf* *in-buf*))

;; Buffers for going in and out of the DB
(defvar *out-buf* (make-buffer-stream))
(defvar *key-buf* (make-buffer-stream))
(defvar *in-buf* (make-buffer-stream))

;; Stuff the serializer uses
(defvar *lisp-obj-id* 0)
(defvar *circularity-hash* (make-hash-table :test 'eq))
#+(or cmu scl sbcl allegro)
(defvar *resourced-byte-spec* (byte 32 0))

;; TODO: make this for real!
(defun run-elephant-thread (thunk)
  (let ((*current-transaction* +NULL-VOID+)
	(*errno-buffer* (allocate-foreign-object :int 1))
	(*get-buffer* (allocate-foreign-object :char 1))
	(*get-buffer-length* 0)
	(*store-controller* *store-controller*)
	(*auto-commit* *auto-commit*)
	(*out-buf* (make-buffer-stream))
	(*key-buf* (make-buffer-stream))
	(*in-buf* (make-buffer-stream))
	(*lisp-obj-id* 0)
	(*circularity-hash* (make-hash-table :test 'eq))
	#+(or cmu scl sbcl allegro)
	(*resourced-byte-spec* (byte 32 0)))
    (declare (special *current-transaction* *errno-buffer*
		      *get-buffer* *get-buffer-length* *store-controller*
		      *auto-commit* *out-buf* *key-buf* *in-buf* 
		      *lisp-obj-id* *circularity-hash* 
		      #+(or cmu scl sbcl allegro) *resourced-byte-spec*))
    (funcall thunk)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Macros

;; Good defaults for elephant
(defmacro with-transaction ((&key transaction 
				  (environment '(controller-environment
						 *store-controller*))
				  (parent '*current-transaction*)
				  dirty-read txn-nosync
				  txn-nowait txn-sync
				  (retries 100))
			    &body body)
  `(sleepycat:with-transaction (:transaction ,transaction
				:environment ,environment
				:parent ,parent
				:dirty-read ,dirty-read
				:txn-nosync ,txn-nosync
				:txn-nowait ,txn-nowait
				:txn-sync ,txn-sync
				:retries ,retries)
    ,@body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Portable value-weak hash-tables for the cache: when the
;;; values are collected, the entries (keys) should be
;;; flushed from the table too

(defun make-cache-table (&rest args)
  #+(or cmu sbcl scl)
  (apply #'make-hash-table args)
  #+allegro
  (apply #'make-hash-table :values :weak args)
  #+lispworks
  (apply #'make-hash-table :weak-kind :value args)
  #-(or cmu sbcl scl allegro lispworks)
  (apply #'make-hash-table args)
  )

(defun get-cache (key cache)
  #+(or cmu sbcl)
  (let ((val (gethash key cache)))
    (if val (values (weak-pointer-value val) t)
	(values nil nil)))
  #-(or cmu sbcl scl)
  (gethash key cache)
  )

(defun make-finalizer (key cache)
  #+(or cmu sbcl)
  (lambda () (remhash key cache))
  #+allegro
  (lambda (obj) (declare (ignore obj)) (remhash key cache))
  )

(defun setf-cache (key cache value)
  #+(or cmu sbcl)
  (let ((w (make-weak-pointer value)))
    (finalize value (make-finalizer key cache))
    (setf (gethash key cache) w)
    value)
  #+allegro
  (progn
    (excl:schedule-finalization value (make-finalizer key cache))
    (setf (gethash key cache) value))
  #-(or cmu sbcl scl allegro)
  (setf (gethash key cache) value)
  )

(defsetf get-cache setf-cache)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; buffer-stream methods

(eval-when (:compile-toplevel :load-toplevel)
  (defun process-struct-slot-defs (slot-defs struct)
    (loop for def in slot-defs
	  collect (list (first def) (list (second def) struct)))))

(defmacro with-struct-slots (slot-defs struct &body body)
  `(symbol-macrolet ,(process-struct-slot-defs slot-defs struct)
    ,@body))

(defun resize-buffer-stream (bs length)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type fixnum length))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (when (> length len)
      (let ((newlen (max length (* len 2))))
	(declare (type fixnum newlen))
	(let ((newbuf (allocate-foreign-object :char newlen)))
	  (copy-bufs newbuf 0 buf 0 len)
	  (free-foreign-object buf)
	  (setf buf newbuf)
	  (setf len newlen)
	  nil)))))

(defun finish-buffer (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position))
    bs
    (let ((length pos))
      (setf pos 0)
      length)))

(defun buffer-write-byte (b bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type (unsigned-byte 8) b))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ pos 1)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (setf (deref-array buf '(:array :char) pos) b)
      (setf pos needed))))

(defun buffer-write-int (i bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type (signed-byte 32) i))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ pos 4)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-int buf i pos)
      (setf pos needed)
      nil)))

(defun buffer-write-uint (u bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type (unsigned-byte 32) u))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ pos 4)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-uint buf u pos)
      (setf pos needed)
      nil)))

(defun buffer-write-float (d bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type single-float d))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ pos 4)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-float buf d pos)
      (setf pos needed)
      nil)))

(defun buffer-write-double (d bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type double-float d))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ pos 8)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-double buf d pos)
      (setf pos needed)
      nil)))

(defun buffer-write-string (s bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type string s))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (pos buffer-stream-position)
		      (len buffer-stream-length))
    bs		      
    (let* ((str-bytes (byte-length s))
	   (needed (+ pos str-bytes)))
      (declare (type fixnum str-bytes needed)
	       (dynamic-extent str-bytes needed))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (copy-str-to-buf buf pos s 0 str-bytes)
      (setf pos needed)
      nil)))

(defun buffer-read-byte (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((pos (buffer-stream-position bs)))
    (incf (buffer-stream-position bs))
    (deref-array (buffer-stream-buffer bs) '(:array :char) pos)))

(defun buffer-read-fixnum (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((pos (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ pos 4))
    (the fixnum (read-int (buffer-stream-buffer bs) pos))))

(defun buffer-read-int (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((pos (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ pos 4))
    (the (signed-byte 32) (read-int (buffer-stream-buffer bs) pos))))

(defun buffer-read-uint (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((pos (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ pos 4))
    (the (unsigned-byte 32)(read-uint (buffer-stream-buffer bs) pos))))

(defun buffer-read-float (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((pos (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ pos 4))
    (read-float (buffer-stream-buffer bs) pos)))

(defun buffer-read-double (bs)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((pos (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ pos 8))
    (read-double (buffer-stream-buffer bs) pos)))

(defun buffer-read-string (bs length)
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type fixnum length))
  (let ((pos (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ pos length))
    ;; wide!!!
    #+(and allegro ics)
    (excl:native-to-string 
     (offset-char-pointer (buffer-stream-buffer bs) pos) 
     :length length
     :external-format :unicode)
    #+lispworks
    (fli:convert-from-foreign-string 
     (offset-char-pointer (buffer-stream-buffer bs) pos)
     :length length :external-format :unicode :null-terminated-p nil)
    #-(or lispworks (and allegro ics))
    (convert-from-foreign-string 
     (offset-char-pointer (buffer-stream-buffer bs) pos) 
     :length length :null-terminated-p nil)))

